#include "acctd-config.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <syslog.h>

#include <string.h>
#include <errno.h>

#include <sys/types.h>
#include <limits.h>

#ifdef HAVE_DB_185_H
#include <db_185.h>
#else
#include <db.h>
#endif

#include <sysexits.h>
#include <fcntl.h>
#include <time.h>
#include <ctype.h>

#ifdef DBF_1
static char lmt_name[] = LMTNAME;
#endif

char *pname;
char *user;			/* user to operate */

DB *db = NULL;
DBT key[1], val[1];
int listonly = 0;

void usage()
{
#ifndef DBF_1

	fprintf(stderr, "\
Usage: %s a|c|m|d|l user [time]\n\
	a <user> <time> - assign <time> minutes for user <user>\n\
	c <user> <time> - create new user <user> with <time> minutes\n\
	m <user> <time> - add/sub <time> minutes for user <user>\n\
	d <user>        - delete user <user>\n\
	l <user>        - list <user> ('-' for list all)\n\
", pname);

#else

    fprintf(stderr, "\
Usage: %s a[z]|c[z]|s[z]|m[z]|p[m]|d|l[z]|t [parm]\n\
 a[z] <user> <time_lim> <time_lim_left> <day_lim> <day_lim_left> [<tot_used>]\n\
                - assign time minutes for user <user>\n\
 c[z] <user> <time_lim> <time_lim_left> <day_lim> <day_lim_left> [<tot_used>]\n\
                - create new user <user> with time minutes\n\
 s[z] <user> <time_lim> <time_lim_left> <day_lim> <day_lim_left> [<tot_used>]\n\
                - set time minutes for user <user>\n\
 m[z] <user> <time_lim> <time_lim_left> <day_lim> <day_lim_left> [<tot_used>]\n\
                - add/sub time minutes for user <user>\n\
 p[m] <user> <parrot_lim> [<parrot_credit>]\n\
 d <user>       - delete user <user>\n\
 l[z] <user>    - list <user> ('-' for list all)\n\
 t [time stamp] - set last modify 'day limits' table\n\
", pname);

#endif
    exit(EX_USAGE);
}

void openbase(void)
{
    if(db)
      return;
    db = dbopen(TAC_PLUS_DBFILE, (listonly ? O_RDONLY : O_RDWR | O_CREAT) | O_EXLOCK, 0600, DB_HASH, NULL);
    if (!db) {
	perror("unable to open user database " TAC_PLUS_DBFILE);
	exit(EX_OSFILE);
    }
}

int adduser(u, create,
#ifdef DBF_1
              tl, tll, dl, dll, tu, tuflag)
char u[];
int create;
long tl;
long tll;
long dl;
long dll;
long tu;
int tuflag;
#else
              tq)
char u[];
int create;
time_t tq;
#endif
{
#ifdef DBF_1
    TQ_st tq;
#endif

    openbase();

    if(create == 'c'
#ifdef DBF_1
       || create == 's' || tuflag < 0
#endif
      )
     {
      key->data = u;
      key->size = strlen(u) + 1;
      (*db->get) (db, key, val, NULL);
#ifdef DBF_1
      if (create == 'c') {
#endif
        if (val->data) {
     	  printf("User `%s' already exists.\n", u);
	  exit(EX_NOUSER);
        }
#ifdef DBF_1
      } else if (create == 's') {
        if (!val->data) {
   	  printf("User `%s' not found.\n", u);
	  exit(EX_NOUSER);
        }
      }
#endif
     }

#ifdef DBF_1
    tq.tl = tl;
    tq.tll = tll;
    tq.dl = dl;
    tq.dll = dll;
    tq.tot = (
              (!tuflag || (tuflag < 0 && !val->data)) ? ((tl >= 0) ? (tl - tll) : (0 - tll)) :
              ((tuflag > 0) ? tu : ((TQ_st *)val->data)->tot)
             );
#ifdef PARROT
    tq.parrot = PARROT_OFF;
    tq.parrot_credit = 0.0;
#endif
#endif

    key->data = u;
    key->size = strlen(u) + 1;
    val->data = (void *) &tq;
    val->size = sizeof(tq);
    return (*db->put) (db, key, val, NULL);
}

int deluser(u)
char u[];
{
    openbase();

    key->data = u;
    key->size = strlen(u) + 1;
    return (*db->del) (db, key, NULL);
}

#ifdef DBF_1

static int ph = 1;

int listprint(u, tq)
char u[];
TQ_st *tq;
{
    if (!strcmp(u, lmt_name))
	return 0;

    if (ph > 0) {
	ph = -1;
	printf("---user---"
#ifdef PARROT
"   --parrots/credit--"
#endif
"  --total time/left--  --day limit/left--  --total online time--\n");
    }
    printf(" %-8s", u);
#ifdef PARROT
    if (tq->parrot == PARROT_OFF)
	printf("          off");
    else
	printf("      %7.2f", tq->parrot);
    printf(" %7.2f", tq->parrot_credit);
#endif
    switch (tq->tl) {
	case TIME_LIMIT_NO:  printf("    nolimit"); break;
	case TIME_LIMIT_OFF: printf("        off"); break;
	default: printf("    %7ld", tq->tl / 60); break;
    }
    printf("   %7ld", tq->tll / 60);
    switch (tq->dl) {
	case TIME_LIMIT_NO:  printf("    nolimit"); break;
	case TIME_LIMIT_OFF: printf("        off"); break;
	default: printf("     %6ld", tq->dl / 60); break;
    }
    printf("   %6ld       %16ld\n", tq->dll / 60, tq->tot / 60);

    return 0;
}

int lmtput(void)
{
    char buf[80];

    if(!ph)
      return 0;

    printf("Last modify 'day limits' table: ");
    key->data = lmt_name;
    key->size = strlen(lmt_name) + 1;
    if ((*db->get) (db, key, val, NULL))
	printf("undefined\n");
    else {
	strftime(buf, sizeof(buf), "%Y/%m/%d %H:%M:%S", localtime((time_t *) val->data));
	printf("%s\n", buf);
    }
    return (EX_NOUSER);
}
#endif

int listuser(u)
char u[];
{
#ifdef DBF_1
    TQ_st *tq;
#endif

    openbase();

    if (strcmp(u, "-")) {
	key->data = u;
	key->size = strlen(u) + 1;
	if (!(*db->get) (db, key, val, NULL)) {
#ifdef DBF_1
	    tq = (TQ_st *) val->data;
	    listprint(u, tq);
#else
            printf("*** User `%s', time left %ld minutes\n",
      	           u, *(time_t *)val->data / 60);
#endif
	} else {
	    printf("*** Time limit not set for user `%s'\n", u);
	    exit(EX_NOUSER);
	}
    } else {
	while(!(*db->seq) (db, key, val, R_NEXT)) {
#ifdef DBF_1
		tq = (TQ_st *) val->data;
		listprint(key->data, tq);
#else
                printf("%-10s %6ld\n", key->data, *(time_t *)val.dptr / 60);
#endif
	}
    }
#ifdef DBF_1
    lmtput();
#endif
    return 0;
}

#ifdef DBF_1
int lmtset(u)
char u[];
{
    time_t t;
    struct tm tm;

    openbase();

    if (u) {
	sscanf(u, "%d/%d/%d %d:%d:%d", &tm.tm_year, &tm.tm_mon, &tm.tm_mday, &tm.tm_hour, &tm.tm_min, &tm.tm_sec);
	tm.tm_year -= 1900;
	tm.tm_mon--;
	tm.tm_isdst = -1;
	t = mktime(&tm);
    } else
	time(&t);
    key->data = lmt_name;
    key->size = strlen(lmt_name) + 1;
    val->data = (void *) &t;
    val->size = sizeof(t);
    return (*db->put) (db, key, val, NULL);
}
#endif

int moduser(u,
#ifdef DBF_1
               tl, tll, dl, dll, tu, tuflag)
char u[];
long tl;
long tll;
long dl;
long dll;
long tu;
int tuflag;
#else
               tq)

char u[];
time_t tq;
#endif
{
#ifdef DBF_1
    TQ_st tq;
    TQ_st *tqp;
#endif

    openbase();

    key->data = u;
    key->size = strlen(u) + 1;
    if ((*db->get) (db, key, val, NULL)) {
	printf("User `%s' not found\n", u);
	exit(EX_NOUSER);
    }
#ifdef DBF_1
    memcpy(&tq, val->data, sizeof(tq));
    tq.tl  += tl;
    tq.tll += tll;
    tq.dl  += dl;
    tq.dll += dll;
    tq.tot = (
              (tuflag <= 0) ?
              ((tuflag)  ?  tq.tot : ((tq.tl >= 0) ? (tq.tl - tq.tll) : (0 - tq.tll))) :
              tq.tot + tu
             );
#else
    tq += *(time_t *)val->data;
#endif
    val->data = (void *) &tq;
    val->size = sizeof(tq);
    return (*db->put) (db, key, val, NULL);
}

#ifdef PARROT
int setparrot(char u[], double parrot, double parrot_credit, int flag) {
    TQ_st tq;
    TQ_st *tqp;

    openbase();

    key->data = u;
    key->size = strlen(u) + 1;
    if ((*db->get) (db, key, val, NULL)) {
	fprintf(stderr, "macct: creating new user %s\n", u);
	adduser(u, 'c', TIME_LIMIT_OFF, 0, TIME_LIMIT_OFF, 0, 0, 0);
    }
    tqp = (TQ_st *) val->data;
    tq.tl  = tqp->tl;
    tq.tll = tqp->tll;
    tq.dl  = tqp->dl;
    tq.dll = tqp->dll;
    tq.tot = tqp->tot;
    tq.parrot = parrot;
    tq.parrot_credit = parrot_credit;
    if (flag) {
	tq.parrot += tqp->parrot;
	tq.parrot_credit += tqp->parrot_credit;
    }
    val->data = (void *) &tq;
    val->size = sizeof(tq);
    return (*db->put) (db, key, val, NULL);
}
#endif

#ifdef DBF_1
int renew_day_limit(void)
{
    DB *dbl;
    TQ_st tq;
    DBT key_l[1], val_l[1];
    char u[17];

    dbl = dbopen(TAC_PLUS_DBFILE, O_RDONLY, 0600, DB_HASH, NULL);
    if (!dbl) {
	perror("unable to open database file (renew) " TAC_PLUS_DBFILE);
	exit(EX_OSFILE);
    }

    while(!(*dbl->seq) (dbl, key_l, val_l, R_NEXT)) {
	    if (strcmp(key_l->data, lmt_name)) {
	      strncpy(u, key_l->data, key_l->size);
	      memcpy(&tq, val_l->data, sizeof(tq));
	      tq.dll = tq.dl;
	      key->data = u;
	      key->size = strlen(u) + 1;
	      val->data = (void *) &tq;
	      val->size = sizeof(tq);
	      (*db->put) (db, key, val, NULL);
	    }
    }
    (*dbl->close) (dbl);
    return 0;
}

#endif

int main(int ac, char *av[]) {
#ifdef DBF_1
    long tu;
    int tuflag;
    long tmp1, tmp2;
#endif
#ifdef PARROT
    double parrot;
    double parrot_credit;
#endif

    pname = strrchr(av[0], '/');
    pname = pname ? ++pname : av[0];

    if((ac == 1) || (*av[1] == 'l'))
      listonly = 1;

    if (ac == 1) {
	listuser("-");
	exit(EX_OK);
    }

    if (ac >= 3) {
      user = av[2];
#ifdef DBF_1
      if (!strcmp(user, lmt_name)) {
	printf("User name `%s' is reserved.\n", user);
	exit(EX_NOUSER);
      }
#endif
    }

    switch (*av[1]) {
    case 'a':
    case 'A':
    case 'c':
    case 'C':
#ifdef DBF_1
    case 's':
    case 'S':
	if (ac != 7 && ac != 8)
#else
	if (ac != 4)
#endif
	    usage();
#ifdef DBF_1
        if(ac == 8)
         {
          tuflag = 1;
          tu = atol(av[7]) * 60;
         }
        else if(*(av[1] + 1) == 'z' || *(av[1] + 1) == 'Z')
         {
          tuflag = -1;
          tu = 0;
         }
        else
         {
          tuflag = 0;
          tu = 0;
         }

	if (!strcasecmp(av[3],"off"))
	    tmp1 = TIME_LIMIT_OFF;
	else if (!strcasecmp(av[3],"nolimit"))
	    tmp1 = TIME_LIMIT_NO;
	else
	    tmp1 = atol(av[3]) * 60;

	if (!strcasecmp(av[5],"off"))
	    tmp2 = TIME_LIMIT_OFF;
	else if (!strcasecmp(av[5],"nolimit"))
	    tmp2 = TIME_LIMIT_NO;
	else
	    tmp2 = atol(av[5]) * 60;

	adduser(user, tolower(*av[1]),
	tmp1, atol(av[4]) * 60, tmp2, atol(av[6]) * 60,
	     tu, tuflag);
#else
	adduser(user, tolower(*av[1]), atoi(av[3]) * 60);
#endif
	break;

    case 'm':
    case 'M':
#ifdef DBF_1
	if (ac != 7 && ac != 8)
#else
	if (ac != 4)
#endif
	    usage();
#ifdef DBF_1
        if(ac == 8)
         {
          tuflag = 1;
          tu = atol(av[7]) * 60;
         }
        else if(*(av[1] + 1) == 'z' || *(av[1] + 1) == 'Z')
         {
          tuflag = -1;
          tu = 0;
         }
        else
         {
          tuflag = 0;
          tu = 0;
         }

	if (!strcasecmp(av[3],"off"))
	    tmp1 = TIME_LIMIT_OFF;
	else if (!strcasecmp(av[3],"nolimit"))
	    tmp1 = TIME_LIMIT_NO;
	else
	    tmp1 = atol(av[3]) * 60;

	if (!strcasecmp(av[5],"off"))
	    tmp2 = TIME_LIMIT_OFF;
	else if (!strcasecmp(av[5],"nolimit"))
	    tmp2 = TIME_LIMIT_NO;
	else
	    tmp2 = atol(av[5]) * 60;

	moduser(user,
	tmp1, atol(av[4]) * 60, tmp2, atol(av[6]) * 60,
	     tu, tuflag);
#else
	moduser(user, atoi(av[3]) * 60);
#endif
	break;

#ifdef PARROT
    case 'p':
    case 'P':
	if (ac != 4 && ac != 5)
	    usage();
	if (!strcasecmp(av[3],"off"))
	    parrot = PARROT_OFF;
	else
	    parrot = atof(av[3]);
        if(ac == 5)
	    parrot_credit = atof(av[4]);
	else
	    parrot_credit = 0.0;
        if(*(av[1] + 1) == 'm' || *(av[1] + 1) == 'M')
	    tuflag = 1;
        else
	    tuflag = 0;

	setparrot(user, parrot, parrot_credit, tuflag);
	break;
#endif

    case 'd':
    case 'D':
	if (ac != 3)
	    usage();
	deluser(user);
	break;
    case 'l':
    case 'L':
	if (ac != 3)
	    usage();
#ifdef DBF_1
        if(*(av[1] + 1) == 'z' || *(av[1] + 1) == 'Z')
          ph = 0;
#endif
	listuser(user);
	break;
#ifdef DBF_1
    case 't':
    case 'T':
	if (ac > 3)
	    usage();
	lmtset(user);
	renew_day_limit();
	break;
#endif
    default:
	usage();
	break;
    }
    if(db)
      (*db->close) (db);
    exit(EX_OK);
}
